""" Generator functions to evaluate trial and test functions

NB: Basis evaluation is only needed for the trial function argument in jacobians, as the
multiplication with the test function is part of the sum factorization kernel.
"""
from dune.perftool.generation import (backend,
                                      domain,
                                      get_counter,
                                      get_global_context_value,
                                      iname,
                                      instruction,
                                      kernel_cached,
                                      temporary_variable,
                                      )
from dune.perftool.sumfact.amatrix import (AMatrix,
                                           basis_functions_per_direction,
                                           name_dtheta,
                                           name_theta,
                                           quadrature_points_per_direction,
                                           )
from dune.perftool.sumfact.sumfact import (setup_theta,
                                           sum_factorization_kernel,
                                           sumfact_iname,
                                           )
from dune.perftool.sumfact.quadrature import quadrature_inames
from dune.perftool.loopy.buffer import initialize_buffer
from dune.perftool.pdelab.driver import FEM_name_mangling
from dune.perftool.pdelab.restriction import restricted_name

from pytools import product

import pymbolic.primitives as prim


def name_sumfact_base_buffer():
    count = get_counter('sumfact_base_buffer')
    name = "buffer_{}".format(str(count))
    return name


@kernel_cached
def sumfact_evaluate_coefficient_gradient(element, name, restriction, component):
    # Get a temporary for the gradient
    from ufl.functionview import select_subelement
    sub_element = select_subelement(element, component)
    rank = len(sub_element.value_shape()) + 1
    shape = sub_element.value_shape() + (element.cell().geometric_dimension(),)
    shape_impl = ('arr',) * rank
    temporary_variable(name, shape=shape, shape_impl=shape_impl)

    # Calculate values with sumfactorization
    theta = name_theta()
    dtheta = name_dtheta()
    rows = quadrature_points_per_direction()
    cols = basis_functions_per_direction()
    theta_matrix = AMatrix(theta, rows, cols)
    dtheta_matrix = AMatrix(dtheta, rows, cols)

    # TODO:
    # - This only covers rank 1
    # - Avoid setting up whole gradient if only one component is needed?
    # Get geometric dimension
    formdata = get_global_context_value('formdata')
    dim = formdata.geometric_dimension
    buffers = []
    for i in range(dim):
        a_matrices = [theta_matrix] * dim
        a_matrices[i] = dtheta_matrix
        a_matrices = tuple(a_matrices)

        # Do stage 1
        buffer_name = name_sumfact_base_buffer()
        initialize_buffer(buffer_name,
                          base_storage_size=product(max(mat.rows, mat.cols) for mat in a_matrices),
                          num=2
                          )

        insn_dep = setup_theta(element, restriction, component, a_matrices, buffer_name)
        var, _ = sum_factorization_kernel(a_matrices,
                                          buffer_name,
                                          insn_dep=frozenset({insn_dep}),
                                          )

        buffers.append(var)

    for i, buf in enumerate(buffers):
        # Write solution from sumfactorization to gradient variable
        from pymbolic.primitives import Subscript, Variable
        from dune.perftool.generation import get_backend
        assignee = Subscript(Variable(name), i)
        expression = Subscript(Variable(buf), tuple(Variable(i) for i in quadrature_inames()))
        instruction(assignee=assignee,
                    expression=expression,
                    forced_iname_deps=frozenset(get_backend("quad_inames")()),
                    forced_iname_deps_is_final=True,
                    )


@kernel_cached
def pymbolic_trialfunction_gradient(element, restriction, component):
    rawname = "gradu" + "_".join(str(c) for c in component)
    name = restricted_name(rawname, restriction)
    from dune.perftool.pdelab.argument import name_coefficientcontainer
    container = name_coefficientcontainer(restriction)
    sumfact_evaluate_coefficient_gradient(element, name, restriction, component)
    from pymbolic.primitives import Variable
    return Variable(name)


@kernel_cached
def pymbolic_trialfunction(element, restriction, component):
    theta = name_theta()
    rows = quadrature_points_per_direction()
    cols = basis_functions_per_direction()
    a_matrix = AMatrix(theta, rows, cols)
    # TODO: this is 2D only
    a_matrices = (a_matrix, a_matrix)
    buffer_name = name_sumfact_base_buffer()

    # Do stage 1
    initialize_buffer(buffer_name,
                      base_storage_size=product(max(mat.rows, mat.cols) for mat in a_matrices),
                      num=2
                      )

    insn_dep = setup_theta(element, restriction, component, a_matrices, buffer_name)
    var, _ = sum_factorization_kernel(a_matrices,
                                      buffer_name,
                                      insn_dep=frozenset({insn_dep}),
                                      )

    return prim.Subscript(prim.Variable(var),
                          tuple(prim.Variable(i) for i in quadrature_inames())
                          )


@iname
def sumfact_lfs_iname(bound, dim):
    name = "sumfac_lfs_{}".format(dim)
    domain(name, bound)
    return name


@backend(interface="lfs_inames", name="sumfact")
def lfs_inames(element, restriction, number=1, context=''):
    assert number == 1
    formdata = get_global_context_value('formdata')
    dim = formdata.geometric_dimension
    return tuple(sumfact_lfs_iname(basis_functions_per_direction(), d) for d in range(dim))


@backend(interface="evaluate_basis")
@kernel_cached
def evaluate_basis(element, name, restriction):
    temporary_variable(name, shape=())
    theta = name_theta()
    quad_inames = quadrature_inames()
    inames = lfs_inames(element, restriction)
    assert(len(quad_inames) == len(inames))

    instruction(expression=prim.Product(tuple(prim.Subscript(prim.Variable(theta),
                                                             (prim.Variable(i), prim.Variable(j))
                                                             )
                                              for (i, j) in zip(quad_inames, inames)
                                              )
                                        ),
                assignee=prim.Variable(name),
                forced_iname_deps=frozenset(quad_inames + inames),
                forced_iname_deps_is_final=True,
                )


def pymbolic_basis(element, restriction, number):
    assert number == 1
    assert element.num_sub_elements() == 0

    name = "phi_{}".format(FEM_name_mangling(element))
    name = restricted_name(name, restriction)
    evaluate_basis(element, name, restriction)

    return prim.Variable(name)


@backend(interface="evaluate_grad")
@kernel_cached
def evaluate_reference_gradient(element, name, restriction):
    from dune.perftool.pdelab.geometry import name_dimension
    temporary_variable(
        name,
        shape=(name_dimension(),))
    quad_inames = quadrature_inames()
    inames = lfs_inames(element, restriction)
    assert(len(quad_inames) == len(inames))

    theta = name_theta()
    dtheta = name_dtheta()

    # Get geometric dimension
    formdata = get_global_context_value('formdata')
    dim = formdata.geometric_dimension

    for i in range(dim):
        calls = [prim.Subscript(prim.Variable(theta), (prim.Variable(m), prim.Variable(n)))
                 for (m, n) in zip(quad_inames, inames)]
        calls[i] = prim.Subscript(prim.Variable(dtheta), (prim.Variable(quad_inames[i]), prim.Variable(inames[i])))
        calls = tuple(calls)

        # assignee = prim.Subscript(prim.Variable(name), tuple(prim.Variable(0)))
        assignee = prim.Subscript(prim.Variable(name), (i,))
        # assignee = prim.Variable(name)
        expression = prim.Product(calls)

        instruction(assignee=assignee,
                    expression=expression,
                    forced_iname_deps=frozenset(quad_inames + inames),
                    forced_iname_deps_is_final=True,
                    )


def pymbolic_reference_gradient(element, restriction, number):
    assert number == 1

    # TODO: Change name?
    name = "js_{}".format(FEM_name_mangling(element))
    name = restricted_name(name, restriction)
    evaluate_reference_gradient(element, name, restriction)

    return prim.Variable(name)
